import multiprocessing
import sys
from pathlib import Path
from typing import List, Mapping, Union

import click
from rich import box
from rich.align import Align
from rich.console import Group
from rich.panel import Panel
from rich.table import Table

from .term import term
from .utils import get_filter_count

SYS_THREADS = multiprocessing.cpu_count()


def make_sample_check_table(
    samples: List[str],
    pipeline: Path,
    output: Path,
    quality: int,
    ratio: int,
    distance: int,
    filter: Mapping[str, Union[int, float]],
    offset: int,
    queue_all: bool,
) -> List[str]:
    """print formatted table for sample queue

    Args:
        samples: Names of the samples.
        pipeline: Directory for all intermediary files.
        output: Directory for final tsv files.
        quality: PHred quality cutoff for filenames.
        ratio: Clustering ratio for starcode.
        distance: Levenstein distance for starcode.
        filter: Dictionary defining how to filter final clustered barcodes.
        offset: Acceptable insertion or deletion from expected length in final sequences.
        queue_all: If true, skip output check and queue all samples.

    """

    processed_samples = []
    table = Table(box=box.SIMPLE, header_style="bold cyan", collapse_padding=True)
    table.add_column(
        "sample",
        justify="center",
        style="green",
        no_wrap=True,
    )
    table.add_column(f"q{quality}", justify="center")
    table.add_column("barcodes", justify="center")
    table.add_column(
        f"r{ratio}d{distance}",
        justify="center",
    )
    if "filter_count" in filter.keys():
        filter_expr = "min(N)"
    else:
        filter_expr = "min(%)"
    table.add_column(f"{filter_expr}_off{offset}", justify="center")

    for sample in samples:
        if not queue_all:
            row = make_row(
                sample, pipeline, output, quality, ratio, distance, filter, offset
            )
            if "[bold green]" in row[0]:
                processed_samples.append(sample)
        else:
            row = [f"[bold yellow]{sample}"] + ["[yellow]Queued"] * 4

        table.add_row(*row)

    term.print(
        Panel.fit(table, title=term.style_title("Queue"), border_style="border"),
        f"\nThere are {len(samples)-len(processed_samples)} samples to finish processing.\n",
    )

    # TODO: seperate table generation from sample checks.
    return processed_samples


def check_in_dir(sample: str, suffix: str, directory: Path) -> str:
    """check the directory for expected file

    Args:
        sample: Name of sample.
        suffix: File name contents following sample.
        directory: Place to look for file.
    Returns:
        Green check if true. Yellow 'Queued' if false.
    """
    filepath = f"{sample}{suffix}"
    if directory / filepath in directory.glob(f"*{suffix}"):
        return "[bold green]\u2713"
    else:
        return "[yellow]Queued"


def make_row(
    sample: str,
    pipeline: Path,
    output: Path,
    quality: int,
    ratio: int,
    distance: int,
    filter: Mapping[str, Union[int, float]],
    offset: int,
) -> List[str]:
    """generate row for sample queue

    Args:
        sample: Name of the sample.
        pipeline: Directory for all intermediary files.
        output: Directory for final tsv files.
        quality: PHred quality cutoff for filenames.
        ratio: Clustering ratio for starcode.
        distance: Levenstein distance for starcode.
        filter: Dictionary defining how to filter final clustered barcodes.
        offset: Acceptable insertion or deletion from expected length in final sequences.
    """

    # start the table row
    row = []
    row.append(check_in_dir(sample, f".q{quality}.fastq", pipeline))
    row.append(check_in_dir(sample, f".q{quality}.barcodes.tsv", pipeline))
    row.append(
        check_in_dir(sample, f".q{quality}.barcodes.r{ratio}d{distance}.tsv", pipeline)
    )
    if row[-1] == "[bold green]\u2713":
        if "filter_percent" in filter:
            filter_count = get_filter_count(
                pipeline / f"{sample}.q{quality}.barcodes.r{ratio}d{distance}.tsv",
                float(filter["filter_percent"]),
            )
        else:
            filter_count = int(filter["filter_count"])
        row.append(
            check_in_dir(
                sample,
                f".q{quality}.barcodes.r{ratio}d{distance}.min{filter_count}_off{offset}.tsv",
                output,
            )
        )

    if set(row[1:]) == {"[bold green]\u2713"}:
        row = [f"[bold green]{sample}"] + row
    else:
        row.extend(["[yellow]Queued"] * (4 - len(row)))
        row = [f"[bold yellow]{sample}"] + row

    return row


def sample_check(
    fastqs: List[Path],
    pipeline: Path,
    output: Path,
    quality: int,
    ratio: int,
    distance: int,
    filter: Mapping[str, Union[int, float]],
    offset: int,
    yes: bool,
) -> List[Path]:
    """check for existence of file outputs for a set of samples

    Args:
        fastqs: List of fastq files.
        pipeline: Directory for all intermediary files.
        output: Directory for final tsv files.
        quality: PHred quality cutoff for filenames.
        ratio: Clustering ratio for starcode.
        distance: Levenstein distance for starcode.
        filter: Dictionary defining how to filter final clustered barcodes.
        offset: Acceptable insertion or deletion from expected length in final sequences.
        yes: If true, skip confirmation check.
    Returns:
        List of fastq files for samples not finished processing.
    """

    samples = {f.name.split(".")[0]: f for f in fastqs}
    queue_all = not pipeline.is_dir()
    processed_samples = make_sample_check_table(
        sorted(samples.keys()),
        pipeline,
        output,
        quality,
        ratio,
        distance,
        filter,
        offset,
        queue_all,
    )

    if not yes and (
        len(processed_samples) != len(samples)
        and not term.confirm("Continue with these samples?")
    ):
        sys.exit()
    if not yes and len(processed_samples) != len(samples):
        term.print()

    return [f for sample, f in samples.items() if sample in processed_samples]


def print_params(ctx: click.Context) -> None:
    """show all parameters prior to run"""

    params = ctx.params
    params.pop("save_config")
    params = {k: v for k, v in params.items() if v is not None}

    grid = Table.grid(expand=True)
    grid.add_column(justify="right", style="bold cyan")
    grid.add_column(justify="center")
    grid.add_column(style="yellow")

    for key, value in params.items():
        grid.add_row(key, ": ", str(value))

    group: Union[Group, Table] = (
        Group(grid, Align(f"[dim]including {ctx.obj['config_file']}", align="center"))
        if "config_file" in ctx.obj
        else grid
    )

    term.print(
        Panel.fit(
            group,
            title=term.style_title(f"{ctx.info_name.capitalize()} Parameters"),  # type: ignore
            border_style="border",
        )
    )

    if "config_file" in ctx.obj:
        grid.add_row("loaded config", ": ", str(ctx.obj["config_file"]))

    if not ctx.info_name == "combine":
        if params["threads"] == 1 and params["threads"] <= SYS_THREADS / 4:
            term.print(
                f"[dim]Only using {params['threads']} of {SYS_THREADS} available threads..."
            )



import shutil
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Union

import click
from click_rich_help import StyledGroup

from ._checks import pre_run_check
from .pycashier import Pycashier
from .term import theme
from .utils import load_params


def init_check(ctx: click.Context, param: str, check: bool) -> None:
    if not check:
        pre_run_check(command=ctx.to_info_dict()["command"]["name"])


def add_options(
    options: List[Callable],
) -> Callable[[click.decorators.FC], click.decorators.FC]:
    def _add_options(func: click.decorators.FC) -> click.decorators.FC:
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


@dataclass
class Option:
    """custom options class to wrap click.option"""

    param_decls: Sequence[str]
    help: str = ""
    default: Optional[Union[str, int, float, Path]] = None
    show_default: bool = False
    is_flag: bool = False
    type: Any = None
    required: bool = False
    params: Dict[str, Any] = field(default_factory=dict)

    def get_click_option(self) -> Callable[[click.decorators.FC], click.decorators.FC]:
        return click.option(
            *self.param_decls,
            help=self.help,
            default=self.default,
            show_default=self.show_default,
            is_flag=self.is_flag,
            type=self.type,
            required=self.required,
            **self.params,
        )

    # def get_click_option_decorator(self, func: click.decorators.FC) ->click.decorators.FC:
    #     func = self.get_click_option()(func)
    #     return func

    def long(self) -> str:
        """return longest param declaration"""
        return max(self.param_decls, key=len)


shared_options = {
    "output": [
        Option(
            ["-o", "--output"],
            help="output directory",
            default="./outs",
            show_default=True,
            type=click.Path(file_okay=False, path_type=Path),
        ),
        Option(
            ["-p", "--pipeline"],
            help="pipeline directory",
            default="./pipeline",
            show_default=True,
            type=click.Path(file_okay=False, path_type=Path),
        ),
    ],
    "general": [
        Option(
            ["-v", "--verbose"],
            help="print the output of underlying software",
            is_flag=True,
        ),
        Option(
            ["-t", "--threads"],
            help="number of cpu cores to use",
            default=1,
            show_default=True,
        ),
        Option(
            ["-c", "--config"],
            help="read parameter values from config file",
            show_default=True,
            default=Path("pycashier.toml"),
            type=click.Path(dir_okay=False),
            params=dict(
                expose_value=False,
                is_eager=True,
                callback=load_params,
            ),
        ),
        Option(
            ["--save-config"],
            help="save current params to file specified by `--config`",
            type=click.Choice(["explicit", "full"], case_sensitive=False),
        ),
        Option(
            ["--skip-init-check"],
            help="skip runtime dependency check",
            is_flag=True,
            params=dict(
                hidden=True,
                expose_value=False,
                callback=init_check,
            ),
        ),
        Option(
            ["-y", "--yes"],
            help="answer yes to prompts",
            is_flag=True,
        ),
    ],
    "trim": [
        Option(
            ["-e", "--error"],
            help="error tolerance supplied to cutadapt",
            default=0.1,
            show_default=True,
        ),
        Option(
            ["-l", "--length"],
            help="target length of extracted barcode",
            default=20,
            show_default=True,
        ),
        Option(
            ["-ua", "--upstream-adapter"],
            help="5' sequence flanking barcode",
            default="ATCTTGTGGAAAGGACGAAACACCG",
        ),
        Option(
            ["-da", "--downstream-adapter"],
            help="3' sequence flanking barcode",
            default="GTTTTAGAGCTAGAAATAGCAAGTT",
        ),
        Option(
            ["--unlinked-adapters"],
            help="run cutadapt using unlinked adapters",
            is_flag=True,
        ),
        Option(
            ["--skip-trimming"],
            help="skip cutadapt trimming entirely and use reads as-is",
            is_flag=True,
        ),
    ],
    "quality": [
        Option(
            ["-q", "--quality"],
            help="minimum PHRED quality for filtering reads",
            default=30,
            show_default=True,
        ),
        Option(
            ["-up", "--unqualified-percent"],
            help="minimum percent of bases which can be below quality threshold",
            default=20,
            show_default=True,
        ),
        Option(
            ["-fa", "--fastp-args"],
            help="additional arguments provided as a string passed verbatim to fastp",
            type=str,
        ),
    ],
    "cluster": [
        Option(
            ["-r", "--ratio"],
            help="ratio to use for message passing clustering",
            default=3,
            show_default=True,
        ),
        Option(
            ["-d", "--distance"],
            help="levenshtein distance for clustering",
            default=1,
            show_default=True,
            type=click.IntRange(1, 8),
        ),
    ],
    "filter": [
        Option(
            ["-fc", "--filter-count"],
            help="minium nominal number of reads",
            type=int,
        ),
        Option(
            ["-fp", "--filter-percent"],
            help="minimum percentage of total reads",
            default=0.005,
            show_default=True,
        ),
        Option(
            ["--offset"],
            help="length offset from target barcode length post-clustering",
            default=1,
            show_default=True,
        ),
    ],
}

options = {
    "merge": [
        Option(
            ["-i", "--input"],
            help="source directory containing gzipped R1 and R2 fastq files",
            required=True,
            type=click.Path(exists=True, file_okay=False, path_type=Path),
        ),
        Option(
            ["-o", "--output"],
            help="output directory",
            default="./mergedfastqs",
            show_default=True,
            type=click.Path(file_okay=False, path_type=Path),
        ),
        shared_options["output"][1],
        *shared_options["general"],
        Option(
            ["-fa", "--fastp-args"],
            help="additional arguments provided as a string passed verbatim to fastp",
            type=str,
        ),
    ],
    "extract": [
        Option(
            ["-i", "--input"],
            help="source directory containing fastq files",
            required=True,
            type=click.Path(exists=True, file_okay=False, path_type=Path),
        ),
        *[opt for opts in shared_options.values() for opt in opts],
    ],
    "scrna": [
        Option(
            ["-i", "--input"],
            help="source directory containing sam files from scRNA-seq",
            required=True,
            type=click.Path(exists=True, file_okay=False, path_type=Path),
        ),
        Option(
            ["-ml", "--minimum-length"],
            help="minimum length of extracted barcode",
            default=10,
            show_default=True,
        ),
        *[
            option
            for option in [
                *shared_options["output"],
                *shared_options["trim"][:-2],
                *shared_options["general"],
            ]
        ],
    ],
    "combine": [
        Option(
            ["-i", "--input"],
            help="source directory containing output files from [hl]pycashier extract[/]",
            default="./outs",
            show_default=True,
            type=click.Path(exists=True, file_okay=False, path_type=Path),
        ),
        Option(
            ["-o", "--output"],
            help="combined tsv of all samples found in input directory",
            default="./combined.tsv",
            show_default=True,
            type=click.Path(exists=False, dir_okay=False, path_type=Path),
        ),
        *[option for option in shared_options["general"][2:-1]],
    ],
}


def get_shared_flags(
    category: str, start: int = 0, end: Optional[int] = None
) -> List[str]:
    end = len(shared_options[category]) if end is None else end
    return [option.long() for option in shared_options[category][start:end]]


_help_groups_ref = {
    "Input/Output Options": ["--input", *get_shared_flags("output")],
    "Trim (Cutadapt) Options": ["--minimum-length", *get_shared_flags("trim")],
    "Cluster (Starcode) Options": get_shared_flags("cluster"),
    "Quality (Fastp) Options": get_shared_flags("quality"),
    "Merge (Fastp) Options": ["--fastp-args"],
    "Filter Options": get_shared_flags("filter"),
    "General Options": [*get_shared_flags("general"), "--help"],
    "hidden": ["--skip-init-check"],
}


def get_help_groups(
    options: List[Option],
    extra_groups: List[str] = [],
) -> Dict[str, List[str]]:
    flags = [option.long() for option in options] + ["--help"]
    groups = ["Input/Output Options", *extra_groups, "General Options"]
    help_groups: Dict[str, List[str]] = {group: [] for group in groups}
    for group in groups:
        for flag in flags:
            if flag in _help_groups_ref["hidden"]:
                continue
            if flag in _help_groups_ref[group]:
                help_groups[group].append(flag)

    return help_groups


CONTEXT_SETTINGS = dict(
    help_option_names=["-h", "--help"],
    max_content_width=shutil.get_terminal_size(fallback=(110, 24))[0],
)


@click.group(
    cls=StyledGroup,
    theme=theme,
    context_settings=CONTEXT_SETTINGS,
)
@click.version_option(package_name="pycashier", prog_name="pycashier")
def cli() -> None:
    """Cash in on DNA Barcode Tags
    \n\n\n
    See `[hl]pycashier COMMAND -h[/]` for more information.
    """
    pass


@cli.command(
    option_groups=get_help_groups(
        options["extract"],
        extra_groups=[
            "Quality (Fastp) Options",
            "Trim (Cutadapt) Options",
            "Cluster (Starcode) Options",
            "Filter Options",
        ],
    ),
    help=Pycashier.extract.__doc__,
)
@add_options([option.get_click_option() for option in options["extract"]])
@click.pass_context
def extract(ctx: click.Context, save_config: bool, **kwargs: Any) -> None:
    pycashier = Pycashier(ctx, save_config)
    pycashier.extract(ctx, **kwargs)


@cli.command(
    option_groups=get_help_groups(
        options["merge"],
        extra_groups=[
            "Quality (Fastp) Options",
            "Merge (Fastp) Options",
        ],
    ),
    help=Pycashier.merge.__doc__,
)
@add_options([option.get_click_option() for option in options["merge"]])
@click.pass_context
def merge(ctx: click.Context, save_config: bool, **kwargs: Any) -> None:
    pycashier = Pycashier(ctx, save_config)
    pycashier.merge(**kwargs)


@cli.command(
    option_groups=get_help_groups(
        options["scrna"],
        extra_groups=[
            "Trim (Cutadapt) Options",
        ],
    ),
    help=Pycashier.scrna.__doc__,
)
@add_options([option.get_click_option() for option in options["scrna"]])
@click.pass_context
def scrna(ctx: click.Context, save_config: bool, **kwargs: Any) -> None:
    pycashier = Pycashier(ctx, save_config)
    pycashier.scrna(**kwargs)


@cli.command(
    option_groups=get_help_groups(options["combine"]),
    help=Pycashier.combine.__doc__,
)
@add_options([option.get_click_option() for option in options["combine"]])
@click.pass_context
def combine(ctx: click.Context, save_config: bool, **kwargs: Any) -> None:
    pycashier = Pycashier(ctx, save_config)
    pycashier.combine(**kwargs)


def main() -> None:
    cli(prog_name="pycashier")


if __name__ == "__main__":
    main()



import sys
import tempfile
from pathlib import Path
from typing import Union

from rich.status import Status

try:
    import pysam
except ImportError:
    pass

from rich.progress import Progress, SpinnerColumn, TimeElapsedColumn

from .term import term
from .utils import check_output, confirm_samples, run_cmd


def sam_to_name_labeled_fastq(sample: str, in_file: Path, out_file: Path) -> None:
    """convert sam file to metadata labeled fastq

    Args:
        sample: Name of sample.
        in_file: Sam file to convert.
        out_file: Converted fastq file.
    """

    new_sam = False

    with open(in_file, "r") as f_in:
        if f_in.readline()[0:3] != "@HD":
            new_sam = True

    sam_file: Union[str, Path]
    if new_sam:
        term.process("sam is headerless, adding a fake one")
        sam_file = fake_header_add(in_file)
    else:
        sam_file = in_file

    sam_length = pysam.AlignmentFile(str(sam_file), "r", check_sq=False).count()
    sam = pysam.AlignmentFile(str(sam_file), "r", check_sq=False)

    with open(out_file, "w") as f_out:

        with Progress(
            SpinnerColumn("pipe", style="line"),
            *Progress.get_default_columns(),
            "Elapsed:",
            TimeElapsedColumn(),
            transient=True,
            console=term._console,
        ) as progress:
            task = progress.add_task("[red]sam -> fastq", total=sam_length)

            for record in sam:

                tagdict = dict(record.tags)
                cell_barcode = None
                if "CB" in tagdict.keys():
                    cell_barcode = tagdict["CB"].split("-")[0]
                elif "CR" in tagdict.keys():
                    cell_barcode = tagdict["CR"]

                umi = None
                if "UB" in tagdict.keys():
                    umi = tagdict["UB"]
                elif "UR" in tagdict.keys():
                    umi = tagdict["UR"]

                if cell_barcode and umi:

                    qualities = record.query_qualities
                    ascii_qualities = "".join([chr(q + 33) for q in qualities])

                    f_out.write(f"@{record.query_name}_{umi}_{cell_barcode}\n")
                    f_out.write(f"{record.query_sequence}\n+\n{ascii_qualities}\n")

                progress.advance(task)

    if new_sam:

        Path(sam_file).unlink()


def labeled_fastq_to_tsv(in_file: Path, out_file: Path) -> None:
    """convert labeled fastq to tsv

    Args:
        in_file: Fastq to convert to tsv.
        out_file: Converted tsv file.
    """

    with open(in_file) as f_in:

        with open(out_file, "w") as f_out:
            read_lines = []
            # TODO: add progress bar
            for line in f_in.readlines():
                read_lines.append(line)
                if len(read_lines) == 4:

                    read_name, umi, cell_barcode = read_lines[0].rstrip("\n").split("_")
                    lineage_barcode = read_lines[1].rstrip("\n")

                    f_out.write(
                        f"{read_name}\t{umi}\
                            \t{cell_barcode}\t{lineage_barcode}\n"
                    )
                    read_lines = []


def fake_header_add(in_file: Path) -> str:

    fake_header = """@HD\tVN:1.6\tSO:coordinate
@SQ\tSN:1\tLN:1000
@SQ\tSN:2\tLN:1000
@SQ\tSN:3\tLN:1000
@SQ\tSN:4\tLN:1000
@SQ\tSN:5\tLN:1000
@SQ\tSN:6\tLN:1000
@SQ\tSN:7\tLN:1000
@SQ\tSN:8\tLN:1000
@SQ\tSN:9\tLN:1000
@SQ\tSN:10\tLN:1000
@SQ\tSN:11\tLN:1000
@SQ\tSN:12\tLN:1000
@SQ\tSN:13\tLN:1000
@SQ\tSN:14\tLN:1000
@SQ\tSN:15\tLN:1000
@SQ\tSN:16\tLN:1000
@SQ\tSN:17\tLN:1000
@SQ\tSN:18\tLN:1000
@SQ\tSN:19\tLN:1000
@SQ\tSN:20\tLN:1000 @SQ\tSN:21\tLN:1000
@SQ\tSN:22\tLN:1000
@SQ\tSN:23\tLN:1000
@SQ\tSN:X\tLN:1000
@SQ\tSN:Y\tLN:1000
@SQ\tSN:MT\tLN:1000
@SQ\tSN:GL000192.1\tLN:1000
@SQ\tSN:GL000225.1\tLN:1000
@SQ\tSN:GL000194.1\tLN:1000
@SQ\tSN:GL000193.1\tLN:1000
@SQ\tSN:GL000200.1\tLN:1000
@SQ\tSN:GL000222.1\tLN:1000
@SQ\tSN:GL000212.1\tLN:1000
@SQ\tSN:GL000195.1\tLN:1000
@SQ\tSN:GL000223.1\tLN:1000
@SQ\tSN:GL000224.1\tLN:1000
@SQ\tSN:GL000219.1\tLN:1000
@SQ\tSN:GL000205.1\tLN:1000
@SQ\tSN:GL000215.1\tLN:1000
@SQ\tSN:GL000216.1\tLN:1000
@SQ\tSN:GL000217.1\tLN:1000
@SQ\tSN:GL000199.1\tLN:1000
@SQ\tSN:GL000211.1\tLN:1000
@SQ\tSN:GL000213.1\tLN:1000
@SQ\tSN:GL000220.1\tLN:1000
@SQ\tSN:GL000218.1\tLN:1000
@SQ\tSN:GL000209.1\tLN:1000
@SQ\tSN:GL000221.1\tLN:1000
@SQ\tSN:GL000214.1\tLN:1000
@SQ\tSN:GL000228.1\tLN:1000
@SQ\tSN:GL000227.1\tLN:1000
@SQ\tSN:GL000191.1\tLN:1000
@SQ\tSN:GL000208.1\tLN:1000
@SQ\tSN:GL000198.1\tLN:1000
@SQ\tSN:GL000204.1\tLN:1000
@SQ\tSN:GL000233.1\tLN:1000
@SQ\tSN:GL000237.1\tLN:1000
@SQ\tSN:GL000230.1\tLN:1000
@SQ\tSN:GL000242.1\tLN:1000
@SQ\tSN:GL000243.1\tLN:1000
@SQ\tSN:GL000241.1\tLN:1000
@SQ\tSN:GL000236.1\tLN:1000
@SQ\tSN:GL000240.1\tLN:1000
@SQ\tSN:GL000206.1\tLN:1000
@SQ\tSN:GL000232.1\tLN:1000
@SQ\tSN:GL000234.1\tLN:1000
@SQ\tSN:GL000202.1\tLN:1000
@SQ\tSN:GL000238.1\tLN:1000
@SQ\tSN:GL000244.1\tLN:1000
@SQ\tSN:GL000248.1\tLN:1000
@SQ\tSN:GL000196.1\tLN:1000
@SQ\tSN:GL000249.1\tLN:1000
@SQ\tSN:GL000246.1\tLN:1000
@SQ\tSN:GL000203.1\tLN:1000
@SQ\tSN:GL000197.1\tLN:1000
@SQ\tSN:GL000245.1\tLN:1000
@SQ\tSN:GL000247.1\tLN:1000
@SQ\tSN:GL000201.1\tLN:1000
@SQ\tSN:GL000235.1\tLN:1000
@SQ\tSN:GL000239.1\tLN:1000
@SQ\tSN:GL000210.1\tLN:1000
@SQ\tSN:GL000231.1\tLN:1000
@SQ\tSN:GL000229.1\tLN:1000
@SQ\tSN:GL000226.1\tLN:1000
"""

    # TODO: change to with statement
    # possible to make this whole function based on decorator maybe?
    f = tempfile.NamedTemporaryFile(delete=False, dir=Path.cwd())
    f.write((bytes(fake_header, encoding="utf-8")))
    f.flush()

    with open(in_file, "r") as sam_file:
        for line in sam_file:
            f.write(bytes(line, encoding="utf-8"))

    f.close()
    return f.name


def single_cell_process(
    sample: str,
    samfile: Path,
    pipeline: Path,
    output: Path,
    error: float,
    minimum_length: int,
    length: int,
    upstream_adapter: str,
    downstream_adapter: str,
    threads: int,
    verbose: bool,
    status: Status,
) -> None:
    """extract barcodes from single cell data

    Args:

        sample: Name of the sample.
        samfile: Path to sam file.
        pipeline: Directory for all intermediary files.
        output: Directory for final tsv files.
        error: Error rate used by cutadapt.
        minimum_length: Minimum barcode length.
        length: Expected lenth of barcode.
        upstream_adapter: 5' barcode flanking sequence.
        downstream_adapter: 3' barcode flanking sequence.
        ratio: Clustering ratio for starcode.
        threads: Number of threads for starcode to use.
        verbose: If true, show subcommand output.
        status: Rich.console status to suspend for stderr printing.
    """

    adapter_string = f"-g {upstream_adapter} -a {downstream_adapter}"
    fastq, barcode_fastq = (
        pipeline / f"{sample}.cell_record_labeled.{ext}"
        for ext in ("fastq", "barcode.fastq")
    )
    tsv_out = output / f"{sample}.cell_record_labeled.barcode.tsv"

    if check_output(fastq, "converting sam to labeled fastq"):
        status.stop()
        sam_to_name_labeled_fastq(sample, samfile, fastq)
        status.start()
    else:
        term.process("skipping sam to labeled fastq conversion")

    if check_output(barcode_fastq, "extracting barcodes w/ [b]cutadapt[/]"):

        command = (
            "cutadapt "
            f"-e {error} "
            f"-j {threads} "
            f"--minimum-length={minimum_length} "
            f"--maximum-length={length} "
            "--max-n=0 "
            f"--trimmed-only {adapter_string} "
            "-n 2 "
            f"-o {barcode_fastq} {fastq}"
        )

        run_cmd(command, sample, barcode_fastq, verbose, status)

    if check_output(tsv_out, "converting labeled fastq to tsv"):
        labeled_fastq_to_tsv(barcode_fastq, tsv_out)


def single_cell(
    input: Path,
    pipeline: Path,
    output: Path,
    error: float,
    minimum_length: int,
    length: int,
    upstream_adapter: str,
    downstream_adapter: str,
    threads: int,
    verbose: bool,
    yes: bool,
) -> None:
    """extract barcodes from set of single cell samples

    Args:

        input: Directory of sam files.
        pipeline: Directory for all intermediary files.
        output: Directory for final tsv files.
        error: Error rate used by cutadapt.
        minimum_length: Minimum barcode length.
        length: Expected lenth of barcode.
        upstream_adapter: 5' barcode flanking sequence.
        downstream_adapter: 3' barcode flanking sequence.
        threads: Number of threads for starcode to use.
        verbose: If true, show subcommand output.
        yes: If true, skip sample confirmation.
    """

    for d in [pipeline, output]:
        d.mkdir(exist_ok=True)

    # TODO: raise error if can't get sample name
    samfiles = {
        f.name.split(".")[0]: f for f in input.iterdir() if not f.name.startswith(".")
    }

    for f in samfiles.values():

        if f.suffix != ".sam":
            term.print(
                f"[ScrnaError]: There is a non sam file in the provided input directory: {f}",
                err=True,
            )
            sys.exit(1)

    confirm_samples(list(samfiles), yes)

    for sample, samfile in samfiles.items():

        term.process(f"[hl]{sample}[/]", status="start")

        with term.cash_in() as status:

            single_cell_process(
                sample,
                samfile,
                pipeline,
                output,
                error,
                minimum_length,
                length,
                upstream_adapter,
                downstream_adapter,
                threads,
                verbose,
                status,
            )

        term.process(status="end")

    term.print("\n[green]FINISHED!")
    sys.exit()



import sys
from shutil import which
from typing import Dict, List

from rich import box
from rich.panel import Panel
from rich.table import Table

from .term import term

PACKAGES = ["cutadapt", "fastp", "starcode", "pysam"]
CMD_PACKAGES: Dict[str, List[str]] = {
    "combine": [],
    "merge": ["fastp"],
    "extract": ["fastp", "cutadapt", "starcode"],
    "scrna": ["pysam", "cutadapt"],
}


def yes_no(condition: bool) -> str:
    """colored yes/no output for table

    Args:
        condition: bool
    Returns:
        Formatted boolean text.
    """

    return "[green]yes[/green]" if condition else "[red]no[/red]"


def pre_run_check(command: str) -> None:
    """Check for runtime dependencies

    Args:
        command: Name of pycashier subcommand.
    """
    pkgs_exist = {name: is_tool(name) for name in PACKAGES}
    if False in pkgs_exist.values():

        table = Table(box=box.SIMPLE, header_style="bold cyan", collapse_padding=True)
        table.add_column("package", justify="center")
        table.add_column("installed", justify="center")
        table.add_column("needed")

        for name, exists in pkgs_exist.items():
            table.add_row(
                name,
                yes_no(exists),
                yes_no(name in CMD_PACKAGES[command]),
                style="bold" if not exists and name in CMD_PACKAGES[command] else "dim",
            )

        term.print(
            f"\n[red bold] FAILED PRE-RUN CHECKS for [hl]pycashier {command}[/hl]!!\n",
            Panel.fit(
                table,
                title="Dependencies",
            ),
            "It's recommended to install pycashier within a conda environment.\n"
            "See the repo for details: [link]https://github.com/brocklab/pycashier[/link]",
        )
        sys.exit(1)


def is_tool(name: str) -> bool:
    """Check whether `name` is on PATH and marked as executable.

    Args:
        name: Executable/dependency.
    Returns:
        True for success, False otherwise
    """
    if name == "pysam":
        try:
            # replace with importlib?
            import pysam  # noqa

            return True
        except ImportError:
            return False
    return which(name) is not None



import shutil
from pathlib import Path
from typing import Dict

from rich.status import Status

from .term import term
from .utils import check_output, fastq_to_tsv, run_cmd


def trim(
    sample: str,
    pipeline: Path,
    fastq: Path,
    quality: int,
    unqualified_percent: float,
    fastp_args: Dict[str, str],
    error: float,
    threads: int,
    length: int,
    distance: int,
    upstream_adapter: str,
    downstream_adapter: str,
    unlinked_adapters: bool,
    skip_trimming: bool,
    verbose: bool,
    status: Status,
) -> None:
    """peform quality filtering and extraction

    Args:
        sample: Name of sample.
        pipeline: Directory for all intermediary files.
        fastq: Path to fastq for sample.
        quality: PHred quality cutoff for filenames.
        unqualified_percent: Percent of bases that may fall below quality cutoff.
        fastp_args: Extra arguments passed to fastp.
        error: Error rate used by cutadapt.
        threads: Number of threads for starcode to use.
        length: Expected lenth of barcode.
        distance: Levenstein distance for starcode.
        upstream_adapter: 5' barcode flanking sequence.
        downstream_adapter: 3' barcode flanking sequence.
        unlinked_adapters: If true use unlinked_adapters with cutadapt.
        skip_trimming: Skip fastp adapter trimming step.
        verbose: If true, show subcommand output.
        status: Rich.console status to suspend for stderr printing.
    """

    json, html = (pipeline / "qc" / f"{sample}.{ext}" for ext in ("json", "html"))
    filtered_fastq, filtered_barcode_fastq = (
        pipeline / f"{sample}.q{quality}.{ext}" for ext in ("fastq", "barcode.fastq")
    )

    adapter_string = (
        f"-g {upstream_adapter} -a {downstream_adapter}"
        if unlinked_adapters
        else f"-g {upstream_adapter}...{downstream_adapter}"
    )

    (pipeline / "qc").mkdir(exist_ok=True)

    if check_output(filtered_fastq, "quality filtering reads w/ [b]fastp[/]"):

        command = (
            "fastp "
            f"-i {fastq} "
            f"-o {filtered_fastq} "
            f"-q {quality} "
            f"-u {unqualified_percent} "
            f"-w {threads} "
            f"-h {html} "
            f"-j {json} "
            "--dont_eval_duplication "
            f"{fastp_args or ''}"
        )

        run_cmd(command, sample, filtered_fastq, verbose, status)

    if skip_trimming:
        shutil.copy(filtered_fastq, filtered_barcode_fastq)

    if check_output(filtered_barcode_fastq, "extracting barcodes w/ [b]cutadapt[/]"):

        command = (
            "cutadapt "
            f"-e {error} "
            f"-j {threads} "
            f"--minimum-length={length - distance} "
            f"--maximum-length={length + distance} "
            f"--max-n=0 "
            f"--trimmed-only "
            f"{adapter_string} "
            "-n 2 "
            f"-o {filtered_barcode_fastq} {filtered_fastq}"
        )

        run_cmd(command, sample, filtered_barcode_fastq, verbose, status)
        term.process("filtering and extraction complete")

    barcodes_out = pipeline / f"{sample}.q{quality}.barcodes.tsv"

    if check_output(barcodes_out, "converting fastq to tsv"):
        fastq_to_tsv(filtered_barcode_fastq, barcodes_out)



from pathlib import Path
from typing import Mapping, Union

from .term import term
from .utils import get_filter_count


def filter_by_percent(
    file_in: Path, filter_percent: float, length: int, offset: int, outdir: Path
) -> None:
    """filter clusted barcodes with nominal abundance cutoff

    Args:
        file_in: Path to clustered barcode counts.
        filter_percent: Minimum percent of total cutoff.
        length: Expected lenth of barcode.
        offset: Acceptable insertion or deletion from expected length in final sequences.
        output: Directory for final tsv files.
    """

    # TODO: docstring

    filter_by_count(
        file_in, get_filter_count(file_in, filter_percent), length, offset, outdir
    )


def filter_by_count(
    file_in: Path, filter_count: int, length: int, offset: int, output: Path
) -> None:
    """filter clusted barcodes with nominal abundance cutoff

    Args:
        file_in: Path to clustered barcode counts.
        filter_count: Minimum count cutoff.
        length: Expected lenth of barcode.
        offset: Acceptable insertion or deletion from expected length in final sequences.
        output: Directory for final tsv files.
    """

    final = output / f"{file_in.stem}.min{filter_count}_off{offset}{file_in.suffix}"

    with open(file_in, "r") as csv_in:
        with open(final, "w") as csv_out:
            for line in csv_in:
                linesplit = line.split("\t")
                if (
                    int(linesplit[1]) >= filter_count
                    and abs(len(linesplit[0]) - length) <= offset
                ):
                    csv_out.write(f"{linesplit[0]}\t{linesplit[1]}")

    if final.stat().st_size == 0:
        term.print(
            "[yellow]WARNING[/]: no barcodes passed final length and abundance filters"
        )


def read_filter(
    sample: str,
    pipeline: Path,
    output: Path,
    length: int,
    offset: int,
    filter: Mapping[str, Union[int, float]],
    quality: int,
    ratio: int,
    distance: int,
) -> None:
    """filter clusted barcodes with final abundance cutoff

    Args:
        sample: Name of the sample.
        pipeline: Directory for all intermediary files.
        output: Directory for final tsv files.
        length: Expected lenth of barcode.
        offset: Acceptable insertion or deletion from expected length in final sequences.
        filter: Dictionary defining how to filter final clustered barcodes.
        quality: PHred quality cutoff for filenames.
        ratio: Clustering ratio for starcode.
        distance: Levenstein distance for starcode.
    """

    file_in = pipeline / f"{sample}.q{quality}.barcodes.r{ratio}d{distance}.tsv"

    if "filter_count" in filter.keys():
        term.process(
            f"post-clustering filtering w/ [b]{filter['filter_count']}[/] read cutoff"
        )

        filter_by_count(file_in, int(filter["filter_count"]), length, offset, output)

    else:
        term.process(
            f"post-clustering filtering w/ [b]{filter['filter_percent']}[/] % cutoff"
        )

        filter_by_percent(file_in, filter["filter_percent"], length, offset, output)



from pathlib import Path
from typing import Any, Dict

import click

from .extract import extract_all
from .merge import merge_all
from .single_cell import single_cell
from .term import term
from .termui import print_params, sample_check
from .utils import combine_outs, get_fastqs, save_params, validate_filter_args


class Pycashier:
    def __init__(self, ctx: click.Context, save_config: bool) -> None:
        if save_config:
            save_params(ctx)
        print_params(ctx)

    def extract(
        self,
        ctx: click.Context,
        input: Path,
        output: Path,
        pipeline: Path,
        quality: int,
        unqualified_percent: float,
        fastp_args: Dict[str, str],
        skip_trimming: bool,
        error: float,
        length: int,
        upstream_adapter: str,
        downstream_adapter: str,
        unlinked_adapters: bool,
        ratio: int,
        distance: int,
        offset: int,
        verbose: bool,
        threads: int,
        yes: bool,
        **kwargs: Any,
    ) -> None:
        """
        extract DNA barcodes from a directory of fastq files

        \b
        Sample names should be delimited with a ".", such as `[hl][yellow]<sample>[/yellow].raw.fastq[/]`,
        anything succeeding the first period will be ignored by `[hl]pycashier[/]`.

        If your data is paired-end with overlapping barcodes, see `[hl]pycashier merge[/]`.
        """

        # validate that filter count and filter percent aren't both defined
        filter = validate_filter_args(ctx)

        term.print(("[b]\n[cyan]PYCASHIER:[/cyan] Starting Extraction\n"))

        fastqs = get_fastqs(input)

        processed_fastqs = sample_check(
            fastqs,
            pipeline,
            output,
            quality,
            ratio,
            distance,
            filter,
            offset,
            yes,
        )

        fastqs = [f for f in fastqs if f not in processed_fastqs]

        extract_all(
            fastqs,
            output,
            pipeline,
            quality,
            unqualified_percent,
            fastp_args,
            skip_trimming,
            error,
            length,
            upstream_adapter,
            downstream_adapter,
            unlinked_adapters,
            ratio,
            distance,
            filter,
            offset,
            verbose,
            threads,
        )

    def merge(
        self,
        input: Path,
        output: Path,
        pipeline: Path,
        fastp_args: Dict[str, str],
        threads: int,
        verbose: bool,
        yes: bool,
    ) -> None:
        """
        merge overlapping paired-end reads using fastp
        \n\n\n
        Simple wrapper over `[hl]fastp[/]` to combine R1 and R2 from PE fastq files.
        \n\n\n
        """

        term.print(("[b]\n[cyan]PYCASHIER:[/cyan] Starting Merge\n"))

        merge_all(
            [f for f in input.iterdir()],
            pipeline,
            output,
            threads,
            verbose,
            fastp_args,
            yes,
        )

    def scrna(
        self,
        input: Path,
        output: Path,
        pipeline: Path,
        minimum_length: int,
        length: int,
        error: float,
        upstream_adapter: str,
        downstream_adapter: str,
        threads: int,
        verbose: bool,
        yes: bool,
    ) -> None:
        """
        extract expressed DNA barcodes from scRNA-seq
        \n
        \b
        Designed for interoperability with 10X scRNA-seq workflow.
        After processing samples with `[hl]cellranger[/]` resulting
        bam files should be converted to sam files using `[hl]samtools[/]`.
        \n
        [i]NOTE[/]: You can speed this up by providing a sam file with only
        the unmapped reads.
        """
        term.print(("[b]\n[cyan]PYCASHIER:[/cyan] Starting Single Cell Extraction\n"))

        single_cell(
            input,
            pipeline,
            output,
            error,
            minimum_length,
            length,
            upstream_adapter,
            downstream_adapter,
            threads,
            verbose,
            yes,
        )

    def combine(
        self,
        input: Path,
        output: Path,
    ) -> None:
        """
        combine resulting output of [hl]extract[/]
        """

        combine_outs(input, output)



import shutil
from textwrap import dedent
from typing import Any, Optional

from rich.console import Console
from rich.highlighter import RegexHighlighter
from rich.panel import Panel
from rich.prompt import Confirm
from rich.status import Status
from rich.text import Text
from rich.theme import Theme

MAX_WIDTH = 110

theme = Theme(
    {
        "hl": "bold cyan",
        "line": "bright_magenta",
        "border": "red",
        # click-rich-help defaults
        "header": "bold italic cyan",
        "option": "bold yellow",
        "metavar": "green",
        "default": "dim",
        "required": "dim red",
    },
    inherit=True,
)


class ErrorHighlighter(RegexHighlighter):
    """Apply style to anything that looks like an error."""

    highlights = [r"(?P<error>\[\w+Error\])"]


class Term:
    """rich-based ui"""

    def __init__(self, width: Optional[int] = None) -> None:
        self._console = Console(highlight=False, theme=theme, width=width)
        self._err_console = Console(
            theme=Theme({"hl": "bold cyan", "error": "bold red"}, inherit=True),
            stderr=True,
            highlighter=ErrorHighlighter(),
            width=width,
        )

    def print(self, *args: Any, err: bool = False, **kwargs: Any) -> None:
        console = self._err_console if err else self._console
        console.print(*args, **kwargs)

    def cash_in(self) -> Status:
        return self._console.status(
            "cashing in...", spinner="pipe", spinner_style="bright_magenta"
        )

    def confirm(self, *args: Any, **kwargs: Any) -> bool:
        return Confirm.ask(*args, console=self._console, **kwargs)

    def style_title(self, title: str) -> Text:
        return Text.from_markup(f"[reset][hl]{title}[/hl][/reset]")

    def textbox(self, text: str, title: str = "", *args: Any, **kwargs: Any) -> None:
        self.print(
            Panel.fit(
                Text(dedent(text), *args, *kwargs),
                title=self.style_title(title),
                title_align="left",
            )
        )

    def subcommand(self, sample: str, pkg: str, command: str, output: str) -> None:
        """print subcommand results"""

        self.print()
        self.textbox(command, title=f"{pkg.upper()} COMMAND | [green]{sample}[/green]")
        self.textbox(output, title=f"{pkg.upper()} OUTPUT | [green]{sample}[/green]")

    def process(self, text: str = "", status: str = "") -> None:
        if status == "start":
            self.print(f"[line]╭── [/]\[{text}]")
        elif status == "end":
            self.print("[line]╰────> [/][green]:heavy_check_mark:[/]")
        elif status == "skip":
            self.print(f"[line]├ [/][dim strike]{Text.from_markup(text)}[/] skipped")
        else:
            self.print(f"[line]├ [/]{text}")


cols = shutil.get_terminal_size().columns
term = Term(width=MAX_WIDTH if cols > MAX_WIDTH else cols)



from pathlib import Path

from rich.status import Status

from .term import term
from .utils import check_output, extract_csv_column, run_cmd


def cluster(
    sample: str,
    pipeline: Path,
    ratio: int,
    distance: int,
    quality: int,
    threads: int,
    verbose: bool,
    status: Status,
) -> None:
    """cluster the barcodes using starcode

    Args:
        sample: Name of the sample.
        pipeline: Directory for all intermediary files.
        ratio: Clustering ratio for starcode.
        distance: Levenstein distance for starcode.
        quality: PHred quality cutoff for filenames.
        threads: Number of threads for starcode to use.
        verbose: If true, show subcommand output.
        status: Console status to suspend for stderr printing.
    """

    extracted_csv = pipeline / f"{sample}.q{quality}.barcodes.tsv"
    clustered = pipeline / f"{sample}.q{quality}.barcodes.r{ratio}d{distance}.tsv"

    if check_output(
        clustered,
        "clustering barcodes w/ [b]starcode[/]",
    ):

        if not extracted_csv.with_suffix(".c2.tsv").is_file():
            # ? if this file exists this wont be reached and pycashier won't know what the "input_file" is
            # TODO: refactor so input_file isn't unbound
            input_file = extract_csv_column(extracted_csv, 2)

        command = f"starcode -d {distance} -r {ratio} -t {threads} -i {input_file} -o {clustered}"

        run_cmd(command, sample, clustered, verbose, status)

        input_file.unlink()

        term.process("clustering complete")



from pathlib import Path
from typing import Dict, List, Mapping, Union

from .cluster import cluster
from .read_filter import read_filter
from .term import term
from .trim import trim


def extract_all(
    fastqs: List[Path],
    output: Path,
    pipeline: Path,
    quality: int,
    unqualified_percent: float,
    fastp_args: Dict[str, str],
    skip_trimming: bool,
    error: float,
    length: int,
    upstream_adapter: str,
    downstream_adapter: str,
    unlinked_adapters: bool,
    ratio: int,
    distance: int,
    filter: Mapping[str, Union[int, float]],
    offset: int,
    verbose: bool,
    threads: int,
) -> None:
    """peform quality filtering, extraction, clustering, and final filtering

    Args:
        fastqs: List of fastq files.
        output: Directory for final tsv files.
        pipeline: Directory for all intermediary files.
        quality: PHred quality cutoff for filenames.
        unqualified_percent: Percent of bases that may fall below quality cutoff.
        fastp_args: Extra arguments passed to fastp.
        skip_trimming: Skip fastp adapter trimming step.
        error: Error rate used by cutadapt.
        length: Expected lenth of barcode.
        upstream_adapter: 5' barcode flanking sequence.
        downstream_adapter: 3' barcode flanking sequence.
        unlinked_adapters: If true use unlinked_adapters with cutadapt.
        ratio: Clustering ratio for starcode.
        distance: Levenstein distance for starcode.
        filter: Dictionary defining how to filter final clustered barcodes.
        offset: Acceptable insertion or deletion from expected length in final sequences.
        verbose: If true, show subcommand output.
        threads: Number of threads for starcode to use.
    """

    for d in [pipeline, output]:
        d.mkdir(exist_ok=True)

    for fastq in sorted(fastqs):

        sample = fastq.name.split(".")[0]

        term.process(f"[hl]{sample}[/]", status="start")

        with term.cash_in() as status:

            trim(
                sample,
                pipeline,
                fastq,
                quality,
                unqualified_percent,
                fastp_args,
                error,
                threads,
                length,
                distance,
                upstream_adapter,
                downstream_adapter,
                unlinked_adapters,
                skip_trimming,
                verbose,
                status,
            )

            cluster(
                sample, pipeline, ratio, distance, quality, threads, verbose, status
            )

            read_filter(
                sample,
                pipeline,
                output,
                length,
                offset,
                filter,
                quality,
                ratio,
                distance,
            )

        term.process(status="end")

    term.print("\n[green]FINISHED!")



import re
import sys
from pathlib import Path
from typing import Dict, List

from rich.status import Status

from .term import term
from .utils import check_output, confirm_samples, run_cmd


def get_pefastqs(fastqs: List[Path]) -> Dict[str, Dict[str, Path]]:
    """parse input files for paired-end sequences

    Args:
        fastqs: list of fastq files from input directory
    Returns:
        dictionary of sample and paired-end fastq files
    """
    pefastqs: Dict[str, Dict[str, Path]] = {}
    for f in fastqs:

        m = re.search(r"(?P<sample>.+?)\..*(?P<read>R[1-2])\..*\.fastq.*", f.name)
        if m:
            sample, read = m.groups()
            pefastqs.setdefault(sample, {})

            if read not in pefastqs[sample]:
                pefastqs[sample][read] = f
            else:
                term.print(
                    f"[MergeError]: detected multiple [hl]{read}[/] files for [hl]{sample}[/]\n"
                    f"files: [b]{f}[/] and [b]{pefastqs[sample][read]}[/]",
                    err=True,
                )
                sys.exit(1)

        else:
            term.print(
                f"[MergeError]: failed to obtain sample/read info from [b]{f}[/]\n",
                "Merge mode expects fastq(.gz) files with R1 or R2 in the name. Exiting.",
                err=True,
            )
            sys.exit(1)

    return pefastqs


def merge_single(
    sample: str,
    pefastqs: Dict[str, Path],
    pipeline: Path,
    output: Path,
    threads: int,
    verbose: bool,
    fastp_args: Dict[str, str],
    status: Status,
) -> None:
    """merge a single sample with fastp

    Args:
        sample: Name of the sample.
        pefasts: Dictionary of paired-end reads and paths.
        pipeline: Directory for all intermediary files.
        output: Directory for final merged fastq files.
        threads: Number of threads for starcode to use.
        verbose: If true, show subcommand output.
        fastp_args: Extra arguments passed to fastp.
        status: Rich.console status to suspend for stderr printing.

    """

    merged_barcode_fastq = output / f"{sample}.merged.raw.fastq"

    if check_output(merged_barcode_fastq, "merging paired end reads w/ [b]fastp[/]"):

        command = (
            "fastp "
            f"-i {pefastqs['R1']}  "
            f"-I {pefastqs['R2']}  "
            f"-w {threads} "
            "-m -c -G -Q -L "
            f"-j {pipeline}/merge_qc/{sample}.json "
            f"-h {pipeline}/merge_qc/{sample}.html "
            f"--merged_out {merged_barcode_fastq} "
            f"{fastp_args or ''}"
        )

        run_cmd(command, sample, merged_barcode_fastq, verbose, status)


def merge_all(
    fastqs: List[Path],
    pipeline: Path,
    output: Path,
    threads: int,
    verbose: bool,
    fastp_args: Dict[str, str],
    yes: bool,
) -> None:
    """iterate over and merge all samples

    Args:
        fastqs: List of all fastq files.
        pipeline: Directory for all intermediary files.
        output: Directory for final merged fastq files.
        threads: Number of threads for starcode to use.
        verbose: If true, show subcommand output.
        fastp_args: Extra arguments passed to fastp.
        yes: If true, skip confirmation check.
    """

    for d in [pipeline, output]:
        d.mkdir(exist_ok=True)

    pefastqs = get_pefastqs(fastqs)

    confirm_samples(list(pefastqs), yes)

    for sample in sorted(pefastqs):
        term.process(f"[hl]{sample}[/]", status="start")

        with term.cash_in() as status:
            merge_single(
                sample,
                pefastqs[sample],
                pipeline,
                output,
                threads,
                verbose,
                fastp_args,
                status,
            )

        term.process("fastqs sucessfully merged")
        term.process(status="end")

    term.print("\n[green]FINISHED!")
    sys.exit()



import csv
import shlex
import subprocess
import sys
from pathlib import Path
from typing import Dict, List

import click
import tomlkit
from rich.status import Status

from .term import term


def get_fastqs(src: Path) -> List[Path]:
    """determine fastq files
    Args:
        src: Input directory that contains fastq files.
    Returns:
        List of fastq files (may be gzipped).
    """

    fastqs = [f for f in src.iterdir() if not f.name.startswith(".")]
    if not fastqs:
        term.print(f"Source dir: {src}, appears to be empty...")
        term.print("Exiting.")
        sys.exit(1)

    for f in fastqs:

        if not any(f.name.endswith(suffix) for suffix in [".fastq", ".fastq.gz"]):
            term.print(
                f"[InputError]: There is a non fastq file in the provided fastq directory: {f}",
                err=True,
            )
            term.print("Exiting.")
            sys.exit(1)

    return fastqs


def fastq_to_tsv(in_file: Path, out_file: Path) -> None:
    """convert fastq file to tsv

    Args:
        in_file: Fastq file to convert.
        out_file: TSV file to write to.
    """

    with open(in_file) as f_in:
        with open(out_file, "w") as f_out:
            lines = f_in.readlines()
            for i, line in enumerate(lines):

                if not line.startswith("@") or lines[i - 1].strip() == "+":
                    continue
                seq_id = line.strip()
                sequence = lines[i + 1].strip()
                f_out.write(f"{seq_id}\t{sequence}\n")


def extract_csv_column(csv_file: Path, column: int) -> Path:
    """get column from csv file

    Args:
        csv_file: File to extract column from.
        column: Column to extract.
    Return:
        File containg only the extracted column.
    """
    ext = csv_file.suffix
    tmp_out = csv_file.with_suffix(f".c{column}{ext}")
    with open(csv_file, "r") as csv_in:
        with open(tmp_out, "w") as csv_out:
            for line in csv_in:
                linesplit = line.split("\t")
                csv_out.write(f"{linesplit[column-1]}")

    return tmp_out


def get_filter_count(file_in: Path, filter_percent: float) -> int:
    """calculate filter cutoff

    Args:
        file_in: Clustered lineage counts to filter.
        filter_percent: Percent cutoff of total reads count.
    Returns:
        Minimum nominal cutoff value.
    """
    total_reads = 0.0

    with open(file_in, newline="") as csvfile:
        spamreader = csv.reader(csvfile, delimiter="\t")
        for row in spamreader:
            total_reads += float(row[1])

    return int(round(total_reads * filter_percent / 100, 0))


def combine_outs(input_dir: Path, output: Path) -> None:
    """combine output tsvs into one file

    Args:
        input_dir: Directory containing csv files to combine.
        output: TSV to save all data to.
    """
    files = {f.name.split(".")[0]: f for f in input_dir.iterdir()}

    term.print(f"Combing output files for {len(files)} samples.")

    with output.open("w") as tsv_out:
        # TODO: make columns configurable?
        tsv_out.write("sample\tsequence\tcount\n")
        for sample, f in files.items():
            with f.open("r") as tsv_in:
                for line in tsv_in:
                    tsv_out.write(f"{sample}\t{line}")


def validate_filter_args(ctx: click.Context) -> Dict[str, float]:
    """validate filter argument from config and CLI

    Args:
        ctx: Click context.
    Returns:
        Dictionary defining the filter type and value.
    """
    if ctx.params["filter_count"] or ctx.params["filter_count"] == 0:
        if ctx.get_parameter_source("filter_percent").value == 3:  # type: ignore
            ctx.params["filter_percent"] = None
            del ctx.params["filter_percent"]
            return {"filter_count": ctx.params["filter_count"]}
        else:
            raise click.BadParameter(
                "`--filter-count` and `--filter-percent` are mutually exclusive"
            )
    else:
        del ctx.params["filter_count"]
        return {"filter_percent": ctx.params["filter_percent"]}


def save_params(ctx: click.Context) -> None:
    """save parameters to config file

    Args:
        ctx: Click context.
    """
    cmd = ctx.info_name
    params = {k: v for k, v in ctx.params.items() if v}
    save_type = params.pop("save_config")

    try:
        config_file = Path(ctx.obj["config_file"])
    except TypeError:
        raise click.BadParameter("use `--save-config` with a specified `--config`")

    if config_file.is_file():
        term.print(f"Updating current config file at [hl]{config_file}")
        with config_file.open("r") as f:
            config = tomlkit.load(f)
    else:
        term.print(f"Staring a config file at [hl]{config_file}")
        config = tomlkit.document()

    if save_type == "explicit":
        params = {k: v for k, v in params.items() if ctx.get_parameter_source(k).value != 3}  # type: ignore

    # sanitize the path's for writing to toml
    for k in ["input", "pipeline", "output"]:
        if k in params.keys():
            params[k] = str(params[k])

    config[cmd] = params  # type: ignore

    null_hints = {"extract": ["filter_count", "fastp_args"], "merge": ["fastp_args"]}
    if cmd in null_hints.keys() and save_type == "full":
        for param in null_hints[cmd]:  # type: ignore
            if param not in params.keys():
                config[cmd].add(tomlkit.comment(f"{param} ="))  # type: ignore

    with config_file.open("w") as f:
        f.write(tomlkit.dumps(config))

    term.print("Exiting...")
    ctx.exit()


def load_params(ctx: click.Context, param: str, filename: Path) -> None:
    """load parameters from config file

    Args:
        ctx: Click context.
        param: Invoked command and config table head.
        filename: Config filename.
    """

    if not filename or ctx.resilient_parsing:
        return

    ctx.default_map = {}
    if Path(filename).is_file():
        with Path(filename).open("r") as f:
            params = tomlkit.load(f)
        if params:
            ctx.default_map = params.get(ctx.info_name, {})
    elif Path(filename) != Path("pycashier.toml"):
        term.print(
            f"[InputError]: Specified config file ({filename}) does not exist.",
            err=True,
        )
        sys.exit(1)

    ctx.obj = {"config_file": filename}


def exit_status(p: subprocess.CompletedProcess, file: Path) -> bool:
    """check command exit status and file size

    Args:
        p: Completed subprocess
        file: File to check for nonzero size.
    Returns:
        True for success, False otherwise.
    """
    return True if p.returncode != 0 or file.stat().st_size == 0 else False


def run_cmd(
    command: str, sample: str, output: Path, verbose: bool, status: Status
) -> None:
    """run a subcommand

    Args:
        command: Subcommand to be run in subprocess.
        sample: Name of sample.
        output: Directory of immediate output.
        verbose: If true, print subcommand output.
        status: Status to pause if writing to stderr printing needed.
    """
    cmd_name = command.split()[0]

    p = subprocess.run(
        shlex.split(command),
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
    )

    # remove 'progress: ##%' output from starcode
    stdout = (
        p.stdout
        if not cmd_name == "starcode"
        else "\n".join(
            [line for line in p.stdout.splitlines() if not line.startswith("progress")]
        )
    )

    if verbose or exit_status(p, output):
        term.subcommand(sample, cmd_name, command, stdout)

    if exit_status(p, output):
        status.stop()
        term.print(
            f"[{cmd_name.capitalize()}Error]: Failed to extract reads for sample: [green]{sample}[/green]\n"
            f"see above for {cmd_name} output",
            err=True,
        )
        sys.exit(1)


def confirm_samples(samples: List[str], yes: bool) -> None:
    """display and confirm samples

    Args:
        samples: List of samples for confirmation.
        yes: True if --yes flag used at runtime.
    """

    term.print(f"[hl]Samples[/]: {', '.join(sorted(samples))}\n")
    if not yes and not term.confirm("Continue with these samples?"):
        sys.exit()
    if not yes:
        term.print()


def check_output(file: Path, message: str) -> bool:
    """check for output file and print message

    Args:
        file: Resulting file of step.
        message: Text to display related to step.
    Returns:
        True for success, False otherwise.
    """

    if not file.is_file():
        term.process(message)
        return True
    else:
        term.process(message, status="skip")
        return False
